package main;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Calendar;

public class Simulation {
	private final static int START_YEAR = 2014;
	private final static int NUMBER_OF_OFFICES = 500;
	private final static double TELECOMMUTING_UPPER_LIMIT = 0.20;
	
	private ArrayList<Worker> workers;
	private ArrayList<Office> offices;
	private ArrayList<Worker> todaysTelecommuters;
	private ArrayList<Worker> todaysCommuters;
	private Highway toWork;
	private Highway fromWork;
	private ArrayList<Emission> emissions;
	private ArrayList<EmissionData> emissionData;
	private ArrayList<EmissionSpecified> emissionSpecified;
	
	private double telecommutingPercentage;
	private int year;
	private ArrayList<Month> months;
	private boolean weekend;
	private String dateFormatted;
	
	public Simulation(){
		initEntities();
		initMonths();
		
		SimpleDateFormat formatter = new SimpleDateFormat("dd-MM-yyyy");
		GregorianCalendar cal = new GregorianCalendar();
		year = START_YEAR;
		while (telecommutingPercentage <= TELECOMMUTING_UPPER_LIMIT && year < START_YEAR+31){
			cal.set(Calendar.YEAR, year);
			int numberOfDays = 365;
			if(cal.isLeapYear(year)){
				numberOfDays++;
			}
			
			for (Office office : offices) {
				office.yearlyConsumption();
				office.flexibleOffice();
			}
			
			for (int dayOfYear = 1; dayOfYear < numberOfDays; dayOfYear++ ){
				cal.set(Calendar.DAY_OF_YEAR, dayOfYear);
				Date date = cal.getTime();
				dateFormatted = formatter.format(date);
				int dayOfWeek = date.getDay();
				weekend = (dayOfWeek == 5 || dayOfWeek == 6) ? true : false;
				Month currentMonth = months.get(date.getMonth());
				currentMonth.setTemperature();
				
				todaysWorkArrangement(dayOfWeek);
				
				home(currentMonth);
				commuting(currentMonth);
				office(currentMonth);
				
				emissions();
				
				cleanUp();
			}
			
			telecommuters();
			if (year >= 2014){
				addMoreTelecommuters();
			}
			
			java.util.Date stop = new java.util.Date();
			System.out.println(new java.sql.Timestamp(stop.getTime()));
			System.out.println(year);
			
			year++;
		}
		
		log();
	}
	
	public void initEntities(){
		workers = new ArrayList<Worker>();
		offices = new ArrayList<Office>();
		emissions = new ArrayList<Emission>();
		emissionData = new ArrayList<EmissionData>();
		emissionSpecified = new ArrayList<EmissionSpecified>();
		
		for (int i = 0; i < NUMBER_OF_OFFICES; i++ ){
			Office office = new Office();
			office.setId(i);
			int numberOfEmployees = office.getNumberOfEmployees();
			for (int j = 0; j < numberOfEmployees; j++){
				Worker worker = new Worker();
				worker.setHome(new Home(worker));
				worker.setCar(new Car(worker));
				worker.setOffice(office);
				workers.add(worker);
				office.addEmployee(worker);
			}
			offices.add(office);
		}
	}
	
	public void initMonths(){
		months = new ArrayList<Month>();
		months.add(new Month("January", 1, 22, 10));
		months.add(new Month("February", 2, 23, 11));
		months.add(new Month("March", 3, 24, 12));
		months.add(new Month("April", 4, 26, 13));
		months.add(new Month("May", 5, 27, 16));
		months.add(new Month("June", 6, 30, 18));
		months.add(new Month("July", 7, 32, 19));
		months.add(new Month("August", 8, 33, 21));
		months.add(new Month("September", 9, 33, 20));
		months.add(new Month("October", 10, 29, 16));
		months.add(new Month("November", 11, 25, 13));
		months.add(new Month("December", 12, 22, 10));
	}
	
	public void todaysWorkArrangement(int day){
		todaysCommuters = new ArrayList<Worker>();
		todaysTelecommuters = new ArrayList<Worker>();
		if(weekend){
			return;
		}
		for (Worker worker : workers) {
			if (worker.getTcDays()[day]){
				todaysTelecommuters.add(worker);
			}
			else {
				todaysCommuters.add(worker);
			}
		}
	}
	
	public void home(Month month){
		for (Worker worker : workers) {
			worker.getHome().dailyConsumption(weekend, month);
			if(todaysTelecommuters.contains(worker)){
				Home home = worker.getHome();
				home.homeOffice(weekend, month);
				home.generalUsageOfElectricity();
				Car car = worker.getCar();
				worker.getEmission().incrementCo2e(car.drive(6.0));
				worker.getEmission().co2eCarAvail = car.drive(6.0);
			}
		}
	}
	
	public void commuting(Month month){
		for (Worker worker : workers) {
			worker.getCar().setEmission(new Emission());
		}
		toWork = new Highway(todaysCommuters, workers.size(), true);
		fromWork = new Highway(todaysCommuters, workers.size(), false);
	}
	
	public void office(Month month){
		for (Office office : offices) {
			office.dailyConsumption(weekend, month);
		}
	}
	
	public void emissions(){
		double co2e = 0.0;
		
		double co2eHomeOffice = 0.0; //worker
		double co2ePersonalDevices = 0.0; //worker
		double co2eHomeGeneralEl = 0.0; //worker
		double co2eOffice = 0.0; //office
		double co2eResidentialRelocation = 0.0; //car
		double co2eCommuting = 0.0; //car
		double co2eCarAvail = 0.0; //worker
		double co2eCongestion = 0.0; //car
		
		EmissionData ed = new EmissionData();
		for (Worker worker : workers) {
			co2e += worker.getHome().getEmission().getCo2e();
		}
		ed.homeEmission = new Emission(co2e, dateFormatted);
		
		co2e = 0.0;
		for (Worker worker : workers) {
			co2e += worker.getCar().getEmission().getCo2e();
			
			co2eCommuting += worker.getCar().getEmission().co2eCommuting;
			co2eResidentialRelocation += worker.getCar().getEmission().co2eResidentialRelocation;
			co2eCongestion += worker.getCar().getEmission().co2eCongestion;
		}
		ed.carEmission = new Emission(co2e, dateFormatted);
		ed.carEmission.co2eCommuting = co2eCommuting;
		ed.carEmission.co2eResidentialRelocation = co2eResidentialRelocation;
		ed.carEmission.co2eCongestion = co2eCongestion;
		
		co2e = 0.0;
		for (Worker worker : workers) {
			co2e += worker.getEmission().getCo2e();
			
			co2eHomeOffice += worker.getEmission().co2eHomeOffice;
			co2ePersonalDevices += worker.getEmission().co2ePersonalDevices;
			co2eHomeGeneralEl += worker.getEmission().co2eHomeGeneralEl;
			co2eCarAvail += worker.getEmission().co2eCarAvail;
		}
		ed.workerEmission = new Emission(co2e, dateFormatted);
		ed.workerEmission.co2eHomeOffice = co2eHomeOffice;
		ed.workerEmission.co2ePersonalDevices = co2ePersonalDevices;
		ed.workerEmission.co2eHomeGeneralEl = co2eHomeGeneralEl;
		ed.workerEmission.co2eCarAvail = co2eCarAvail;
		
		co2e = 0.0;
		for (Office office : offices) {
			co2e += office.getEmission().getCo2e();
			
			co2eOffice += office.getEmission().co2eOffice;
		}
		ed.officeEmission = new Emission(co2e, dateFormatted);
		ed.officeEmission.co2eOffice = co2eOffice;
		emissionData.add(ed);
	}
	
	public void addMoreTelecommuters(){
		double newTelecommutingPercentage = sCurve(year-START_YEAR);
		double incrementalIncrease = newTelecommutingPercentage-telecommutingPercentage;
		telecommutingPercentage = newTelecommutingPercentage;
		int numberOfNewTelecommuters = (int) (incrementalIncrease*workers.size());
		commuterToTelecommuter(numberOfNewTelecommuters);
	}
	
	public double sCurve(int year){
		double c1 = -3.8194;
		double c2 = 0.2218;
		double eExp = Math.exp(c1+c2*year);
		return (TELECOMMUTING_UPPER_LIMIT*eExp)/(1+eExp);
	}
	
	public void commuterToTelecommuter(int numberOfNewTelecommuters){
		int newTelecommuters = 0;
		
		while (newTelecommuters < numberOfNewTelecommuters){
			Worker worker = workers.get((int) ((workers.size()-1)*Start.random.nextDouble()));
			if (!worker.isTelecommuter()){
				worker.becomeTelecommuter();
				worker.setTcYear(year);
				newTelecommuters++;
			}
		}
	}
	
	public void telecommuters(){
		for (Worker worker : workers) {
			if(worker.isTelecommuter()){
				int yearsOfTc = year - worker.getTcYear();
				Home home = worker.getHome();
				
				// DEVICES
				if ((yearsOfTc % 3) == 0){
					home.equipPhone();
				}
				if ((yearsOfTc % 5) == 0){
					home.equipComputer();
				}
				
				// RELOCATE
				if (yearsOfTc == 5){
					double random = Start.random.nextDouble();
					if (random < .25){
						home.setHasMoved(true);
						double newDistanceToOffice = home.getDistanceToOffice()*1.834;
						home.setDistanceToOffice(newDistanceToOffice);
						double newM2 = home.getM2()*1.25;
						home.setM2(newM2);
					}
				}
			}
		}
	}
	
	public void cleanUp(){
		todaysTelecommuters = null;
		todaysCommuters = null;
		for (Worker worker : workers) {
			worker.setEmission(new Emission());
		}
	}
	
	public void log(){
		System.out.println(workers.size());
		Log log = new Log(workers);
//		log.setEmissions(emissions);
//		log.dailyCo2e();
//		System.out.println("done");
		log.emissionData = this.emissionData;
		log.dailyCo2e();
		log.monthlyCo2e();
		log.yearlyCo2e();
		log.weeklyCo2e();
		log.reboundBlocksCo2e();
	}
}
